//Copyright (c) 1999, Art Gittleman
//This example is provided WITHOUT ANY WARRANTY either expressed or implied.

/* Uses wait and notify to enable
 * producer and consumer threads to
 * cooperate in using a buffer.
 */

// Updated to concurrent by Dr. Spiegel
//  Note commented lines.
//   Also, each producer now produces endlessly, 
//   and each consumer consumes endlessly
  
public class ConcurrentPutGet {

  class Buffer {
    int buffer[];             
    int size;        // buffer size
    int putpos=0;
    int getpos=0;
    int number=0;

    public Buffer(int size)
    { this.size=size;
      buffer = new int [size]; 
    }

    public synchronized void put(int value) throws InterruptedException {
//    if (number==size) {
      while (number==size) {
        System.out.println("Cannot put -- Buffer full");
        wait();
      }
      number++;
      buffer[putpos] = value;
      System.out.println("Put "+value);
      putpos=(putpos+1)%size;
//    if (number == 1) 
	notify();         
    }
    public synchronized int get() throws InterruptedException {
//    if (number==0) {
      while (number==0) {
         System.out.println("Buffer empty");
         wait();
      }
      number--;
      int n = buffer[getpos];
      System.out.println("Get "+n);
      getpos=(getpos+1)%size;
//    if (number==size-1) 
	notify();    
      return n;
    }
  }

  class Producer extends Thread {
     Buffer buf;

     Producer(Buffer b) { 
        buf = b;
     }			

     public void run() {
       for(int i=1; ; i++)
         try {
             buf.put(i);
	     sleep(100);
           }catch (InterruptedException e){
              e.printStackTrace();
           }
     }
  }

  class Consumer extends Thread {
     Buffer buf;

     Consumer(Buffer b) { 
       buf = b;
     }

     public void run() {
//       for (int i=1; i<=10;i++)
      while (true)
         try {
           buf.get();
	   sleep(100);
         }catch (InterruptedException e) {
               e.printStackTrace();
         }
     }
  }

  public static void main(String[] args) {
    ConcurrentPutGet pg = new ConcurrentPutGet();
    Buffer b = pg.new Buffer(Integer.parseInt(args[0]));
    if (args.length!=3) {
	System.out.println("Usage: PutGet  <Buffer Size> <#Producers>  <#Consumers>");
        System.exit(0);
    }
    Producer p[] = new Producer[Integer.parseInt(args[1])];
    for (int pIdx=0;pIdx<Integer.parseInt(args[1]);pIdx++) {
      p[pIdx]=pg.new Producer(b);
      p[pIdx].start();
    }
    Consumer c[] = new Consumer[Integer.parseInt(args[2])];
    for (int cIdx=0;cIdx<Integer.parseInt(args[2]);cIdx++) {
      c[cIdx]=pg.new Consumer(b);
      c[cIdx].start();
    }
  }
}

